#cython: language_level=3
cimport write
cimport read
cimport prepare
import cython

{%  if quantlane %}
class DictWrapper(dict):
    '''
    This is used only in read_union to be able
    to add `schema` attribute to a dict.
    '''
{% endif %}



@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
def __invoke():
    {{ code | indent(4) }}
    return locals()